@@ -2,18 +2,18 @@
 
 const Long = require('mongodb-core').BSON.Long;
 const MongoError = require('mongodb-core').MongoError;
+const ObjectID = require('mongodb-core').BSON.ObjectID;
+const BSON = require('mongodb-core').BSON;
+const MongoWriteConcernError = require('mongodb-core').MongoWriteConcernError;
 const toError = require('../utils').toError;
 const handleCallback = require('../utils').handleCallback;
 const applyRetryableWrites = require('../utils').applyRetryableWrites;
 const applyWriteConcern = require('../utils').applyWriteConcern;
-const ObjectID = require('mongodb-core').BSON.ObjectID;
-const BSON = require('mongodb-core').BSON;
+const executeOperation = require('../utils').executeOperation;
+const isPromiseLike = require('../utils').isPromiseLike;
 
 // Error codes
-const UNKNOWN_ERROR = 8;
-const INVALID_BSON_ERROR = 22;
 const WRITE_CONCERN_ERROR = 64;
-const MULTIPLE_ERROR = 65;
 
 // Insert types
 const INSERT = 1;
@@ -54,18 +54,6 @@ class Batch {
   }
 }
 
-/**
- * Wraps a legacy operation so we can correctly rewrite it's error
- * @ignore
- */
-class LegacyOp {
-  constructor(batchType, operation, index) {
-    this.batchType = batchType;
-    this.index = index;
-    this.operation = operation;
-  }
-}
-
 /**
  * Create a new BulkWriteResult instance (INTERNAL TYPE, do not instantiate directly)
  *
@@ -352,7 +340,7 @@ class WriteError {
  * Merges results into shared data structure
  * @ignore
  */
-function mergeBatchResults(ordered, batch, bulkResult, err, result) {
+function mergeBatchResults(batch, bulkResult, err, result) {
   // If we have an error set the result to be the err object
   if (err) {
     result = err;
@@ -484,6 +472,41 @@ function mergeBatchResults(ordered, batch, bulkResult, err, result) {
   }
 }
 
+function executeCommands(bulkOperation, options, callback) {
+  if (bulkOperation.s.batches.length === 0) {
+    return handleCallback(callback, null, new BulkWriteResult(bulkOperation.s.bulkResult));
+  }
+
+  const batch = bulkOperation.s.batches.shift();
+
+  function resultHandler(err, result) {
+    // Error is a driver related error not a bulk op error, terminate
+    if (((err && err.driver) || (err && err.message)) && !(err instanceof MongoWriteConcernError)) {
+      return handleCallback(callback, err);
+    }
+
+    // If we have and error
+    if (err) err.ok = 0;
+    if (err instanceof MongoWriteConcernError) {
+      return handleMongoWriteConcernError(batch, bulkOperation.s.bulkResult, err, callback);
+    }
+
+    // Merge the results together
+    const writeResult = new BulkWriteResult(bulkOperation.s.bulkResult);
+    const mergeResult = mergeBatchResults(batch, bulkOperation.s.bulkResult, err, result);
+    if (mergeResult != null) {
+      return handleCallback(callback, null, writeResult);
+    }
+
+    if (bulkOperation.handleWriteError(callback, writeResult)) return;
+
+    // Execute the next command in line
+    executeCommands(bulkOperation, options, callback);
+  }
+
+  bulkOperation.finalOptionsHandler({ options, batch, resultHandler }, callback);
+}
+
 /**
  * handles write concern error
  *
@@ -493,8 +516,8 @@ function mergeBatchResults(ordered, batch, bulkResult, err, result) {
  * @param {WriteConcernError} err
  * @param {function} callback
  */
-function handleMongoWriteConcernError(batch, bulkResult, ordered, err, callback) {
-  mergeBatchResults(ordered, batch, bulkResult, null, err.result);
+function handleMongoWriteConcernError(batch, bulkResult, err, callback) {
+  mergeBatchResults(batch, bulkResult, null, err.result);
 
   const wrappedWriteConcernError = new WriteConcernError({
     errmsg: err.result.writeConcernError.errmsg,
@@ -906,6 +929,15 @@ class BulkOperationBase {
     );
   }
 
+  _handleEarlyError(err, callback) {
+    if (typeof callback === 'function') {
+      callback(err, null);
+      return;
+    }
+
+    return this.s.promiseLibrary.reject(err);
+  }
+
   /**
    * Execute next write command in a chain
    *
@@ -919,19 +951,17 @@ class BulkOperationBase {
     if (typeof options === 'function') (callback = options), (options = {});
     options = options || {};
 
-    if (this.s.executed) {
-      const executedError = toError('batch cannot be re-executed');
-      return typeof callback === 'function'
-        ? callback(executedError, null)
-        : this.s.promiseLibrary.reject(executedError);
-    }
-
     if (typeof _writeConcern === 'function') {
       callback = _writeConcern;
     } else if (_writeConcern && typeof _writeConcern === 'object') {
       this.s.writeConcern = _writeConcern;
     }
 
+    if (this.s.executed) {
+      const executedError = toError('batch cannot be re-executed');
+      return this._handleEarlyError(executedError, callback);
+    }
+
     // If we have current batch
     if (this.isOrdered) {
       if (this.s.currentBatch) this.s.batches.push(this.s.currentBatch);
@@ -943,13 +973,43 @@ class BulkOperationBase {
     // If we have no operations in the bulk raise an error
     if (this.s.batches.length === 0) {
       const emptyBatchError = toError('Invalid Operation, no operations specified');
-      return typeof callback === 'function'
-        ? callback(emptyBatchError, null)
-        : this.s.promiseLibrary.reject(emptyBatchError);
+      return this._handleEarlyError(emptyBatchError, callback);
     }
     return { options, callback };
   }
 
+  /**
+   * The callback format for results
+   * @callback BulkOperationBase~resultCallback
+   * @param {MongoError} error An error instance representing the error during the execution.
+   * @param {BulkWriteResult} result The bulk write result.
+   */
+  /**
+   * Execute the ordered bulk operation
+   *
+   * @method
+   * @param {object} [options] Optional settings.
+   * @param {(number|string)} [options.w] The write concern.
+   * @param {number} [options.wtimeout] The write concern timeout.
+   * @param {boolean} [options.j=false] Specify a journal write concern.
+   * @param {boolean} [options.fsync=false] Specify a file sync write concern.
+   * @param {BulkOperationBase~resultCallback} [callback] The result callback
+   * @throws {MongoError} Throws error if the bulk object has already been executed
+   * @throws {MongoError} Throws error if the bulk object does not have any operations
+   * @return {Promise} returns Promise if no callback passed
+   */
+  execute(_writeConcern, options, callback) {
+    const ret = this.bulkExecute(_writeConcern, options, callback);
+    if (!ret || isPromiseLike(ret)) {
+      return ret;
+    }
+
+    options = ret.options;
+    callback = ret.callback;
+
+    return executeOperation(this.s.topology, executeCommands, [this, options, callback]);
+  }
+
   /**
    * Handles final options before executing command
    *
@@ -1033,11 +1093,7 @@ class BulkOperationBase {
       // Force top level error
       err.ok = 0;
       // Merge top level error and return
-      handleCallback(
-        callback,
-        null,
-        mergeBatchResults(false, config.batch, this.s.bulkResult, err, null)
-      );
+      handleCallback(callback, null, mergeBatchResults(config.batch, this.s.bulkResult, err, null));
     }
   }
 
@@ -1094,20 +1150,8 @@ Object.defineProperty(BulkOperationBase.prototype, 'length', {
 module.exports = {
   Batch,
   BulkOperationBase,
-  BulkWriteError,
-  BulkWriteResult,
   bson,
-  FindOperators,
-  handleMongoWriteConcernError,
-  LegacyOp,
-  mergeBatchResults,
-  INVALID_BSON_ERROR: INVALID_BSON_ERROR,
-  MULTIPLE_ERROR: MULTIPLE_ERROR,
-  UNKNOWN_ERROR: UNKNOWN_ERROR,
-  WRITE_CONCERN_ERROR: WRITE_CONCERN_ERROR,
   INSERT: INSERT,
   UPDATE: UPDATE,
-  REMOVE: REMOVE,
-  WriteError,
-  WriteConcernError
+  REMOVE: REMOVE
 };
