@@ -4,7 +4,6 @@ const deprecate = require('util').deprecate;
 const deprecateOptions = require('./utils').deprecateOptions;
 const checkCollectionName = require('./utils').checkCollectionName;
 const ObjectID = require('mongodb-core').BSON.ObjectID;
-const AggregationCursor = require('./aggregation_cursor');
 const MongoError = require('mongodb-core').MongoError;
 const toError = require('./utils').toError;
 const normalizeHintField = require('./utils').normalizeHintField;
@@ -19,10 +18,10 @@ const unordered = require('./bulk/unordered');
 const ordered = require('./bulk/ordered');
 const ChangeStream = require('./change_stream');
 const executeOperation = require('./utils').executeOperation;
-const applyWriteConcern = require('./utils').applyWriteConcern;
 const resolveReadPreference = require('./utils').resolveReadPreference;
 
 // Operations
+const aggregate = require('./operations/aggregate').aggregate;
 const bulkWrite = require('./operations/collection_ops').bulkWrite;
 const checkForAtomicOperators = require('./operations/collection_ops').checkForAtomicOperators;
 const count = require('./operations/collection_ops').count;
@@ -46,12 +45,12 @@ const group = require('./operations/collection_ops').group;
 const indexes = require('./operations/collection_ops').indexes;
 const indexExists = require('./operations/collection_ops').indexExists;
 const indexInformation = require('./operations/collection_ops').indexInformation;
+const insertMany = require('./operations/collection_ops').insertMany;
 const insertOne = require('./operations/collection_ops').insertOne;
 const isCapped = require('./operations/collection_ops').isCapped;
 const mapReduce = require('./operations/collection_ops').mapReduce;
 const optionsOp = require('./operations/collection_ops').optionsOp;
 const parallelCollectionScan = require('./operations/collection_ops').parallelCollectionScan;
-const prepareDocs = require('./operations/collection_ops').prepareDocs;
 const reIndex = require('./operations/collection_ops').reIndex;
 const removeDocuments = require('./operations/collection_ops').removeDocuments;
 const rename = require('./operations/collection_ops').rename;
@@ -269,7 +268,7 @@ const DEPRECATED_FIND_OPTIONS = ['maxScan', 'fields', 'snapshot'];
  * @param {boolean} [options.promoteBuffers=false] Promotes Binary BSON values to native Node Buffers.
  * @param {(ReadPreference|string)} [options.readPreference] The preferred read preference (ReadPreference.PRIMARY, ReadPreference.PRIMARY_PREFERRED, ReadPreference.SECONDARY, ReadPreference.SECONDARY_PREFERRED, ReadPreference.NEAREST).
  * @param {boolean} [options.partial=false] Specify if the cursor should return partial results when querying against a sharded system
- * @param {number} [options.maxTimeMS] Number of miliseconds to wait before aborting the query.
+ * @param {number} [options.maxTimeMS] Number of milliseconds to wait before aborting the query.
  * @param {object} [options.collation] Specify collation (MongoDB 3.4 or higher) settings for update operation (see 3.4 documentation for available fields).
  * @param {ClientSession} [options.session] optional session to use for this operation
  * @throws {MongoError}
@@ -455,7 +454,7 @@ Collection.prototype.insertOne = function(doc, options, callback) {
   if (typeof options === 'function') (callback = options), (options = {});
   options = options || {};
 
-  // Add ignoreUndfined
+  // Add ignoreUndefined
   if (this.s.options.ignoreUndefined) {
     options = Object.assign({}, options);
     options.ignoreUndefined = this.s.options.ignoreUndefined;
@@ -464,21 +463,6 @@ Collection.prototype.insertOne = function(doc, options, callback) {
   return executeOperation(this.s.topology, insertOne, [this, doc, options, callback]);
 };
 
-function mapInsertManyResults(docs, r) {
-  const finalResult = {
-    result: { ok: 1, n: r.insertedCount },
-    ops: docs,
-    insertedCount: r.insertedCount,
-    insertedIds: r.insertedIds
-  };
-
-  if (r.getLastOp()) {
-    finalResult.result.opTime = r.getLastOp();
-  }
-
-  return finalResult;
-}
-
 /**
  * Inserts an array of documents into MongoDB. If documents passed in do not contain the **_id** field,
  * one will be added to each of the documents missing it by the driver, mutating the document. This behavior
@@ -502,33 +486,7 @@ Collection.prototype.insertMany = function(docs, options, callback) {
   if (typeof options === 'function') (callback = options), (options = {});
   options = options ? Object.assign({}, options) : { ordered: true };
 
-  if (!Array.isArray(docs) && typeof callback === 'function') {
-    return callback(
-      MongoError.create({ message: 'docs parameter must be an array of documents', driver: true })
-    );
-  } else if (!Array.isArray(docs)) {
-    return new this.s.promiseLibrary((resolve, reject) => {
-      reject(
-        MongoError.create({ message: 'docs parameter must be an array of documents', driver: true })
-      );
-    });
-  }
-
-  // If keep going set unordered
-  options['serializeFunctions'] = options['serializeFunctions'] || this.s.serializeFunctions;
-
-  docs = prepareDocs(this, docs, options);
-
-  // Generate the bulk write operations
-  const operations = [
-    {
-      insertMany: docs
-    }
-  ];
-
-  return executeOperation(this.s.topology, bulkWrite, [this, operations, options, callback], {
-    resultMutator: result => mapInsertManyResults(docs, result)
-  });
+  return executeOperation(this.s.topology, insertMany, [this, docs, options, callback]);
 };
 
 /**
@@ -727,7 +685,7 @@ Collection.prototype.updateOne = function(filter, update, options, callback) {
 
   options = Object.assign({}, options);
 
-  // Add ignoreUndfined
+  // Add ignoreUndefined
   if (this.s.options.ignoreUndefined) {
     options = Object.assign({}, options);
     options.ignoreUndefined = this.s.options.ignoreUndefined;
@@ -755,7 +713,7 @@ Collection.prototype.replaceOne = function(filter, doc, options, callback) {
   if (typeof options === 'function') (callback = options), (options = {});
   options = Object.assign({}, options);
 
-  // Add ignoreUndfined
+  // Add ignoreUndefined
   if (this.s.options.ignoreUndefined) {
     options = Object.assign({}, options);
     options.ignoreUndefined = this.s.options.ignoreUndefined;
@@ -791,7 +749,7 @@ Collection.prototype.updateMany = function(filter, update, options, callback) {
 
   options = Object.assign({}, options);
 
-  // Add ignoreUndfined
+  // Add ignoreUndefined
   if (this.s.options.ignoreUndefined) {
     options = Object.assign({}, options);
     options.ignoreUndefined = this.s.options.ignoreUndefined;
@@ -824,7 +782,7 @@ Collection.prototype.update = deprecate(function(selector, update, options, call
   if (typeof options === 'function') (callback = options), (options = {});
   options = options || {};
 
-  // Add ignoreUndfined
+  // Add ignoreUndefined
   if (this.s.options.ignoreUndefined) {
     options = Object.assign({}, options);
     options.ignoreUndefined = this.s.options.ignoreUndefined;
@@ -871,7 +829,7 @@ Collection.prototype.deleteOne = function(filter, options, callback) {
   if (typeof options === 'function') (callback = options), (options = {});
   options = Object.assign({}, options);
 
-  // Add ignoreUndfined
+  // Add ignoreUndefined
   if (this.s.options.ignoreUndefined) {
     options = Object.assign({}, options);
     options.ignoreUndefined = this.s.options.ignoreUndefined;
@@ -898,7 +856,7 @@ Collection.prototype.deleteMany = function(filter, options, callback) {
   if (typeof options === 'function') (callback = options), (options = {});
   options = Object.assign({}, options);
 
-  // Add ignoreUndfined
+  // Add ignoreUndefined
   if (this.s.options.ignoreUndefined) {
     options = Object.assign({}, options);
     options.ignoreUndefined = this.s.options.ignoreUndefined;
@@ -927,7 +885,7 @@ Collection.prototype.remove = deprecate(function(selector, options, callback) {
   if (typeof options === 'function') (callback = options), (options = {});
   options = options || {};
 
-  // Add ignoreUndfined
+  // Add ignoreUndefined
   if (this.s.options.ignoreUndefined) {
     options = Object.assign({}, options);
     options.ignoreUndefined = this.s.options.ignoreUndefined;
@@ -954,7 +912,7 @@ Collection.prototype.save = deprecate(function(doc, options, callback) {
   if (typeof options === 'function') (callback = options), (options = {});
   options = options || {};
 
-  // Add ignoreUndfined
+  // Add ignoreUndefined
   if (this.s.options.ignoreUndefined) {
     options = Object.assign({}, options);
     options.ignoreUndefined = this.s.options.ignoreUndefined;
@@ -1005,7 +963,7 @@ Collection.prototype.save = deprecate(function(doc, options, callback) {
  * @param {boolean} [options.promoteBuffers=false] Promotes Binary BSON values to native Node Buffers.
  * @param {(ReadPreference|string)} [options.readPreference] The preferred read preference (ReadPreference.PRIMARY, ReadPreference.PRIMARY_PREFERRED, ReadPreference.SECONDARY, ReadPreference.SECONDARY_PREFERRED, ReadPreference.NEAREST).
  * @param {boolean} [options.partial=false] Specify if the cursor should return partial results when querying against a sharded system
- * @param {number} [options.maxTimeMS] Number of miliseconds to wait before aborting the query.
+ * @param {number} [options.maxTimeMS] Number of milliseconds to wait before aborting the query.
  * @param {object} [options.collation] Specify collation (MongoDB 3.4 or higher) settings for update operation (see 3.4 documentation for available fields).
  * @param {ClientSession} [options.session] optional session to use for this operation
  * @param {Collection~resultCallback} [callback] The command result callback
@@ -1168,7 +1126,7 @@ Collection.prototype.createIndexes = function(indexSpecs, options, callback) {
  * @param {number} [options.wtimeout] The write concern timeout.
  * @param {boolean} [options.j=false] Specify a journal write concern.
  * @param {ClientSession} [options.session] optional session to use for this operation
- * @param {number} [options.maxTimeMS] Number of miliseconds to wait before aborting the query.
+ * @param {number} [options.maxTimeMS] Number of milliseconds to wait before aborting the query.
  * @param {Collection~resultCallback} [callback] The command result callback
  * @return {Promise} returns Promise if no callback passed
  */
@@ -1188,7 +1146,7 @@ Collection.prototype.dropIndex = function(indexName, options, callback) {
  * @method
  * @param {Object} [options] Optional settings
  * @param {ClientSession} [options.session] optional session to use for this operation
- * @param {number} [options.maxTimeMS] Number of miliseconds to wait before aborting the query.
+ * @param {number} [options.maxTimeMS] Number of milliseconds to wait before aborting the query.
  * @param {Collection~resultCallback} [callback] The command result callback
  * @return {Promise} returns Promise if no callback passed
  */
@@ -1360,7 +1318,7 @@ Collection.prototype.indexInformation = function(options, callback) {
  * @param {boolean} [options.skip] The number of documents to skip for the count.
  * @param {string} [options.hint] An index name hint for the query.
  * @param {(ReadPreference|string)} [options.readPreference] The preferred read preference (ReadPreference.PRIMARY, ReadPreference.PRIMARY_PREFERRED, ReadPreference.SECONDARY, ReadPreference.SECONDARY_PREFERRED, ReadPreference.NEAREST).
- * @param {number} [options.maxTimeMS] Number of miliseconds to wait before aborting the query.
+ * @param {number} [options.maxTimeMS] Number of milliseconds to wait before aborting the query.
  * @param {ClientSession} [options.session] optional session to use for this operation
  * @param {Collection~countCallback} [callback] The command result callback
  * @return {Promise} returns Promise if no callback passed
@@ -1434,13 +1392,13 @@ Collection.prototype.countDocuments = function(query, options, callback) {
 };
 
 /**
- * The distinct command returns returns a list of distinct values for the given key across a collection.
+ * The distinct command returns a list of distinct values for the given key across a collection.
  * @method
  * @param {string} key Field of the document to find distinct values for.
  * @param {object} query The query for filtering the set of documents to which we apply the distinct filter.
  * @param {object} [options] Optional settings.
  * @param {(ReadPreference|string)} [options.readPreference] The preferred read preference (ReadPreference.PRIMARY, ReadPreference.PRIMARY_PREFERRED, ReadPreference.SECONDARY, ReadPreference.SECONDARY_PREFERRED, ReadPreference.NEAREST).
- * @param {number} [options.maxTimeMS] Number of miliseconds to wait before aborting the query.
+ * @param {number} [options.maxTimeMS] Number of milliseconds to wait before aborting the query.
  * @param {ClientSession} [options.session] optional session to use for this operation
  * @param {Collection~resultCallback} [callback] The command result callback
  * @return {Promise} returns Promise if no callback passed
@@ -1557,6 +1515,13 @@ Collection.prototype.findOneAndReplace = function(filter, replacement, options,
   if (replacement == null || typeof replacement !== 'object')
     throw toError('replacement parameter must be an object');
 
+  // Check that there are no atomic operators
+  const keys = Object.keys(replacement);
+
+  if (keys[0] && keys[0][0] === '$') {
+    throw toError('The replacement document must not contain atomic operators.');
+  }
+
   return executeOperation(this.s.topology, findOneAndReplace, [
     this,
     filter,
@@ -1629,7 +1594,18 @@ Collection.prototype.findOneAndUpdate = function(filter, update, options, callba
  * @return {Promise} returns Promise if no callback passed
  * @deprecated use findOneAndUpdate, findOneAndReplace or findOneAndDelete instead
  */
-Collection.prototype.findAndModify = deprecate(function(query, sort, doc, options, callback) {
+Collection.prototype.findAndModify = deprecate(
+  _findAndModify,
+  'collection.findAndModify is deprecated. Use findOneAndUpdate, findOneAndReplace or findOneAndDelete instead.'
+);
+
+/**
+ * @ignore
+ */
+
+Collection.prototype._findAndModify = _findAndModify;
+
+function _findAndModify(query, sort, doc, options, callback) {
   const args = Array.prototype.slice.call(arguments, 1);
   callback = typeof args[args.length - 1] === 'function' ? args.pop() : undefined;
   sort = args.length ? args.shift() || [] : [];
@@ -1649,7 +1625,7 @@ Collection.prototype.findAndModify = deprecate(function(query, sort, doc, option
     options,
     callback
   ]);
-}, 'collection.findAndModify is deprecated. Use findOneAndUpdate, findOneAndReplace or findOneAndDelete instead.');
+}
 
 /**
  * Find and remove a document.
@@ -1733,97 +1709,7 @@ Collection.prototype.aggregate = function(pipeline, options, callback) {
     pipeline = args;
   }
 
-  // Ignore readConcern option
-  let ignoreReadConcern = false;
-
-  // Build the command
-  const command = { aggregate: this.s.name, pipeline: pipeline };
-
-  // If out was specified
-  if (typeof options.out === 'string') {
-    pipeline.push({ $out: options.out });
-    // Ignore read concern
-    ignoreReadConcern = true;
-  } else if (pipeline.length > 0 && pipeline[pipeline.length - 1]['$out']) {
-    ignoreReadConcern = true;
-  }
-
-  // Decorate command with writeConcern if out has been specified
-  if (
-    pipeline.length > 0 &&
-    pipeline[pipeline.length - 1]['$out'] &&
-    this.s.topology.capabilities().commandsTakeWriteConcern
-  ) {
-    applyWriteConcern(command, { db: this.s.db, collection: this }, options);
-  }
-
-  // Have we specified collation
-  try {
-    decorateWithCollation(command, this, options);
-  } catch (err) {
-    if (typeof callback === 'function') return callback(err, null);
-    throw err;
-  }
-
-  // If we have bypassDocumentValidation set
-  if (options.bypassDocumentValidation === true) {
-    command.bypassDocumentValidation = options.bypassDocumentValidation;
-  }
-
-  // Do we have a readConcern specified
-  if (!ignoreReadConcern) {
-    decorateWithReadConcern(command, this, options);
-  }
-
-  // If we have allowDiskUse defined
-  if (options.allowDiskUse) command.allowDiskUse = options.allowDiskUse;
-  if (typeof options.maxTimeMS === 'number') command.maxTimeMS = options.maxTimeMS;
-
-  // If we are giving a hint
-  if (options.hint) command.hint = options.hint;
-
-  options = Object.assign({}, options);
-  // Ensure we have the right read preference inheritance
-  options.readPreference = resolveReadPreference(options, { db: this.s.db, collection: this });
-
-  // If explain has been specified add it
-  if (options.explain) {
-    if (command.readConcern || command.writeConcern) {
-      throw toError('"explain" cannot be used on an aggregate call with readConcern/writeConcern');
-    }
-    command.explain = options.explain;
-  }
-
-  if (typeof options.comment === 'string') command.comment = options.comment;
-
-  // Validate that cursor options is valid
-  if (options.cursor != null && typeof options.cursor !== 'object') {
-    throw toError('cursor options must be an object');
-  }
-
-  options.cursor = options.cursor || {};
-  if (options.batchSize) options.cursor.batchSize = options.batchSize;
-  command.cursor = options.cursor;
-
-  // promiseLibrary
-  options.promiseLibrary = this.s.promiseLibrary;
-
-  // Set the AggregationCursor constructor
-  options.cursorFactory = AggregationCursor;
-  if (typeof callback !== 'function') {
-    if (!this.s.topology.capabilities()) {
-      throw new MongoError('cannot connect to server');
-    }
-
-    // Allow disk usage command
-    if (typeof options.allowDiskUse === 'boolean') command.allowDiskUse = options.allowDiskUse;
-    if (typeof options.maxTimeMS === 'number') command.maxTimeMS = options.maxTimeMS;
-
-    // Execute the cursor
-    return this.s.topology.cursor(this.s.namespace, command, options);
-  }
-
-  return handleCallback(callback, null, this.s.topology.cursor(this.s.namespace, command, options));
+  return aggregate(this.s.db, this, pipeline, options, callback);
 };
 
 /**
@@ -1874,7 +1760,7 @@ Collection.prototype.watch = function(pipeline, options) {
  * @param {Collection~parallelCollectionScanCallback} [callback] The command result callback
  * @return {Promise} returns Promise if no callback passed
  */
-Collection.prototype.parallelCollectionScan = function(options, callback) {
+Collection.prototype.parallelCollectionScan = deprecate(function(options, callback) {
   if (typeof options === 'function') (callback = options), (options = { numCursors: 1 });
   // Set number of cursors to 1
   options.numCursors = options.numCursors || 1;
@@ -1894,7 +1780,7 @@ Collection.prototype.parallelCollectionScan = function(options, callback) {
   return executeOperation(this.s.topology, parallelCollectionScan, [this, options, callback], {
     skipSessions: true
   });
-};
+}, 'parallelCollectionScan is deprecated in MongoDB v4.1');
 
 /**
  * Execute a geo search using a geo haystack index on a collection.
