@@ -23,6 +23,7 @@ const CONSTANTS = require('./constants');
 
 // Operations
 const addUser = require('./operations/db_ops').addUser;
+const aggregate = require('./operations/aggregate').aggregate;
 const collections = require('./operations/db_ops').collections;
 const createCollection = require('./operations/db_ops').createCollection;
 const createIndex = require('./operations/db_ops').createIndex;
@@ -112,7 +113,7 @@ const legalOptionNames = [
  * @param {object} [options.pkFactory] A primary key factory object for generation of custom _id keys.
  * @param {object} [options.promiseLibrary] A Promise library class the application wishes to use such as Bluebird, must be ES6 compatible
  * @param {object} [options.readConcern] Specify a read concern for the collection. (only MongoDB 3.2 or higher supported)
- * @param {object} [options.readConcern.level='local'] Specify a read concern level for the collection operations, one of [local|majority]. (only MongoDB 3.2 or higher supported)
+ * @param {ReadConcernLevel} [options.readConcern.level='local'] Specify a read concern level for the collection operations (only MongoDB 3.2 or higher supported)
  * @property {(Server|ReplSet|Mongos)} serverConfig Get the current db topology.
  * @property {number} bufferMaxEntries Current bufferMaxEntries value for the database
  * @property {string} databaseName The name of the database this instance represents.
@@ -263,6 +264,44 @@ Db.prototype.command = function(command, options, callback) {
   return executeOperation(this.s.topology, executeCommand, [this, command, options, callback]);
 };
 
+/**
+ * Execute an aggregation framework pipeline against the database, needs MongoDB >= 3.6
+ * @method
+ * @param {object} [pipeline=[]] Array containing all the aggregation framework commands for the execution.
+ * @param {object} [options] Optional settings.
+ * @param {(ReadPreference|string)} [options.readPreference] The preferred read preference (ReadPreference.PRIMARY, ReadPreference.PRIMARY_PREFERRED, ReadPreference.SECONDARY, ReadPreference.SECONDARY_PREFERRED, ReadPreference.NEAREST).
+ * @param {object} [options.cursor] Return the query as cursor, on 2.6 > it returns as a real cursor on pre 2.6 it returns as an emulated cursor.
+ * @param {number} [options.cursor.batchSize] The batchSize for the cursor
+ * @param {boolean} [options.explain=false] Explain returns the aggregation execution plan (requires mongodb 2.6 >).
+ * @param {boolean} [options.allowDiskUse=false] allowDiskUse lets the server know if it can use disk to store temporary results for the aggregation (requires mongodb 2.6 >).
+ * @param {number} [options.maxTimeMS] maxTimeMS specifies a cumulative time limit in milliseconds for processing operations on the cursor. MongoDB interrupts the operation at the earliest following interrupt point.
+ * @param {boolean} [options.bypassDocumentValidation=false] Allow driver to bypass schema validation in MongoDB 3.2 or higher.
+ * @param {boolean} [options.raw=false] Return document results as raw BSON buffers.
+ * @param {boolean} [options.promoteLongs=true] Promotes Long values to number if they fit inside the 53 bits resolution.
+ * @param {boolean} [options.promoteValues=true] Promotes BSON values to native types where possible, set to false to only receive wrapper types.
+ * @param {boolean} [options.promoteBuffers=false] Promotes Binary BSON values to native Node Buffers.
+ * @param {object} [options.collation] Specify collation (MongoDB 3.4 or higher) settings for update operation (see 3.4 documentation for available fields).
+ * @param {string} [options.comment] Add a comment to an aggregation command
+ * @param {string|object} [options.hint] Add an index selection hint to an aggregation command
+ * @param {ClientSession} [options.session] optional session to use for this operation
+ * @param {Database~aggregationCallback} callback The command result callback
+ * @return {(null|AggregationCursor)}
+ */
+Db.prototype.aggregate = function(pipeline, options, callback) {
+  if (typeof options === 'function') {
+    callback = options;
+    options = {};
+  }
+
+  // If we have no options or callback we are doing
+  // a cursor based aggregation
+  if (options == null && callback == null) {
+    options = {};
+  }
+
+  return aggregate(this, '1', pipeline, options, callback);
+};
+
 /**
  * Return the Admin db instance
  * @method
@@ -281,6 +320,13 @@ Db.prototype.admin = function() {
  * @param {Collection} collection The collection instance.
  */
 
+/**
+ * The callback format for an aggregation call
+ * @callback Database~aggregationCallback
+ * @param {MongoError} error An error instance representing the error during the execution.
+ * @param {AggregationCursor} cursor The cursor if the aggregation command was executed successfully.
+ */
+
 const collectionKeys = [
   'pkFactory',
   'readPreference',
@@ -309,7 +355,7 @@ const collectionKeys = [
  * @param {boolean} [options.serializeFunctions=false] Serialize functions on any object.
  * @param {boolean} [options.strict=false] Returns an error if the collection does not exist
  * @param {object} [options.readConcern] Specify a read concern for the collection. (only MongoDB 3.2 or higher supported)
- * @param {object} [options.readConcern.level='local'] Specify a read concern level for the collection operations, one of [local|majority]. (only MongoDB 3.2 or higher supported)
+ * @param {ReadConcernLevel} [options.readConcern.level='local'] Specify a read concern level for the collection operations (only MongoDB 3.2 or higher supported)
  * @param {Db~collectionResultCallback} [callback] The collection result callback
  * @return {Collection} return the new Collection instance if not in strict mode
  */
@@ -549,7 +595,7 @@ Db.prototype.listCollections = function(filter, options) {
  * @param {Code} code JavaScript to execute on server.
  * @param {(object|array)} parameters The parameters for the call.
  * @param {object} [options] Optional settings.
- * @param {boolean} [options.nolock=false] Tell MongoDB not to block on the evaulation of the javascript.
+ * @param {boolean} [options.nolock=false] Tell MongoDB not to block on the evaluation of the javascript.
  * @param {ClientSession} [options.session] optional session to use for this operation
  * @param {Db~resultCallback} [callback] The results callback
  * @deprecated Eval is deprecated on MongoDB 3.2 and forward
@@ -821,7 +867,7 @@ Db.prototype.setProfilingLevel = function(level, options, callback) {
 };
 
 /**
- * Retrive the current profiling information for MongoDB
+ * Retrieve the current profiling information for MongoDB
  *
  * @param {Object} [options] Optional settings
  * @param {ClientSession} [options.session] optional session to use for this operation
