@@ -97,10 +97,10 @@ function addUser(db, username, password, options, callback) {
       const userPassword = md5.digest('hex');
 
       // If we have another db set
-      const db = options.dbName ? new Db(options.dbName, db.s.topology, db.s.options) : db;
+      const dbToUse = options.dbName ? new Db(options.dbName, db.s.topology, db.s.options) : db;
 
       // Fetch a user collection
-      const collection = db.collection(CONSTANTS.SYSTEM_USER_COLLECTION);
+      const collection = dbToUse.collection(CONSTANTS.SYSTEM_USER_COLLECTION);
 
       // Check if we are inserting the first user
       count(collection, {}, finalOptions, (err, count) => {
@@ -601,7 +601,7 @@ function listCollectionsTransforms(databaseName) {
 }
 
 /**
- * Retrive the current profiling information for MongoDB
+ * Retrieve the current profiling information for MongoDB
  *
  * @method
  * @param {Db} db The Db instance on which to retrieve the profiling info.
