@@ -1,6 +1,5 @@
 'use strict';
 
-const authenticate = require('../authenticate');
 const deprecate = require('util').deprecate;
 const Logger = require('mongodb-core').Logger;
 const MongoError = require('mongodb-core').MongoError;
@@ -10,6 +9,8 @@ const ReadPreference = require('mongodb-core').ReadPreference;
 const ReplSet = require('../topologies/replset');
 const Server = require('../topologies/server');
 const ServerSessionPool = require('mongodb-core').Sessions.ServerSessionPool;
+const NativeTopology = require('../topologies/native_topology');
+const MongoCredentials = require('mongodb-core').MongoCredentials;
 
 let client;
 function loadClient() {
@@ -108,7 +109,10 @@ const validOptionNames = [
   'minSize',
   'monitorCommands',
   'retryWrites',
-  'useNewUrlParser'
+  'useNewUrlParser',
+  'useUnifiedTopology',
+  'serverSelectionTimeoutMS',
+  'useRecoveryToken'
 ];
 
 function addListeners(mongoClient, topology) {
@@ -125,7 +129,10 @@ function addListeners(mongoClient, topology) {
 
 function assignTopology(client, topology) {
   client.topology = topology;
-  topology.s.sessionPool = new ServerSessionPool(topology.s.coreTopology);
+  topology.s.sessionPool =
+    topology instanceof NativeTopology
+      ? new ServerSessionPool(topology)
+      : new ServerSessionPool(topology.s.coreTopology);
 }
 
 // Clear out all events
@@ -174,7 +181,7 @@ function connect(mongoClient, url, options, callback) {
     throw new Error('no callback function provided');
   }
 
-  // Get a logger for MongoClient
+  let didRequestAuthentication = false;
   const logger = Logger('MongoClient', options);
 
   // Did we pass in a Server/ReplSet/Mongos
@@ -211,29 +218,34 @@ function connect(mongoClient, url, options, callback) {
       return callback(new Error('connection string must contain at least one seed host'));
     }
 
+    if (_finalOptions.auth && !_finalOptions.credentials) {
+      try {
+        didRequestAuthentication = true;
+        _finalOptions.credentials = generateCredentials(
+          mongoClient,
+          _finalOptions.auth.user,
+          _finalOptions.auth.password,
+          _finalOptions
+        );
+      } catch (err) {
+        return callback(err);
+      }
+    }
+
+    if (_finalOptions.useUnifiedTopology) {
+      return createTopology(mongoClient, 'unified', _finalOptions, connectCallback);
+    }
+
     // Do we have a replicaset then skip discovery and go straight to connectivity
     if (_finalOptions.replicaSet || _finalOptions.rs_name) {
-      return createTopology(
-        mongoClient,
-        'replicaset',
-        _finalOptions,
-        connectHandler(mongoClient, _finalOptions, connectCallback)
-      );
+      return createTopology(mongoClient, 'replicaset', _finalOptions, connectCallback);
     } else if (object.servers.length > 1) {
-      return createTopology(
-        mongoClient,
-        'mongos',
-        _finalOptions,
-        connectHandler(mongoClient, _finalOptions, connectCallback)
-      );
+      return createTopology(mongoClient, 'mongos', _finalOptions, connectCallback);
     } else {
-      return createServer(
-        mongoClient,
-        _finalOptions,
-        connectHandler(mongoClient, _finalOptions, connectCallback)
-      );
+      return createServer(mongoClient, _finalOptions, connectCallback);
     }
   });
+
   function connectCallback(err, topology) {
     const warningMessage = `seed list contains no mongos proxies, replicaset connections requires the parameter replicaSet to be supplied in the URI or options object, mongodb://server:port/db?replicaSet=name`;
     if (err && err.message === 'no mongos proxies found in seed list') {
@@ -245,35 +257,15 @@ function connect(mongoClient, url, options, callback) {
       return callback(new MongoError(warningMessage));
     }
 
+    if (didRequestAuthentication) {
+      mongoClient.emit('authenticated', null, true);
+    }
+
     // Return the error and db instance
     callback(err, topology);
   }
 }
 
-function connectHandler(client, options, callback) {
-  return (err, topology) => {
-    if (err) {
-      return handleConnectCallback(err, topology, callback);
-    }
-
-    // No authentication just reconnect
-    if (!options.auth) {
-      return handleConnectCallback(err, topology, callback);
-    }
-
-    // Authenticate
-    authenticate(client, options.user, options.password, options, (err, success) => {
-      if (success) {
-        handleConnectCallback(null, topology, callback);
-      } else {
-        if (topology) topology.close();
-        const authError = err ? err : new Error('Could not authenticate user ' + options.auth[0]);
-        handleConnectCallback(authError, topology, callback);
-      }
-    });
-  };
-}
-
 /**
  * Connect to MongoDB using a url as documented at
  *
@@ -314,27 +306,21 @@ function connectWithUrl(mongoClient, url, options, connectCallback) {
     );
   }
 
-  // Connect
-  return url.connect(
-    finalOptions,
-    connectHandler(mongoClient, finalOptions, (err, topology) => {
-      if (err) return connectCallback(err, topology);
-      if (finalOptions.user || finalOptions.password || finalOptions.authMechanism) {
-        return authenticate(
-          mongoClient,
-          finalOptions.user,
-          finalOptions.password,
-          finalOptions,
-          err => {
-            if (err) return connectCallback(err, topology);
-            connectCallback(err, topology);
-          }
-        );
-      }
+  const isDoingAuth = finalOptions.user || finalOptions.password || finalOptions.authMechanism;
+  if (isDoingAuth && !finalOptions.credentials) {
+    try {
+      finalOptions.credentials = generateCredentials(
+        mongoClient,
+        finalOptions.user,
+        finalOptions.password,
+        finalOptions
+      );
+    } catch (err) {
+      return connectCallback(err, url);
+    }
+  }
 
-      connectCallback(err, topology);
-    })
-  );
+  return url.connect(finalOptions, connectCallback);
 }
 
 function createListener(mongoClient, event) {
@@ -410,6 +396,8 @@ function createTopology(mongoClient, topologyType, options, callback) {
     topology = new Mongos(servers, options);
   } else if (topologyType === 'replicaset') {
     topology = new ReplSet(servers, options);
+  } else if (topologyType === 'unified') {
+    topology = new NativeTopology(options.servers, options);
   }
 
   // Add listeners
@@ -465,36 +453,10 @@ function createUnifiedOptions(finalOptions, options) {
   return finalOptions;
 }
 
-function handleConnectCallback(err, topology, callback) {
-  return process.nextTick(() => {
-    try {
-      callback(err, topology);
-    } catch (err) {
-      if (topology) topology.close();
-      throw err;
-    }
-  });
-}
-
 function legacyTransformUrlOptions(object) {
   return mergeOptions(createUnifiedOptions({}, object), object, false);
 }
 
-/**
- * Logout user from server, fire off on all connections and remove all auth info.
- *
- * @method
- * @param {MongoClient} mongoClient The MongoClient instance on which to logout.
- * @param {object} [options] Optional settings. See MongoClient.prototype.logout for a list of options.
- * @param {Db~resultCallback} [callback] The command result callback
- */
-function logout(mongoClient, dbName, callback) {
-  mongoClient.topology.logout(dbName, err => {
-    if (err) return callback(err);
-    callback(null, true);
-  });
-}
-
 function mergeOptions(target, source, flatten) {
   for (const name in source) {
     if (source[name] && typeof source[name] === 'object' && flatten) {
@@ -556,21 +518,17 @@ function transformUrlOptions(_object) {
       object[camelCaseName] = object[name];
     }
   }
-  if (_object.auth) {
-    const auth = _object.auth;
-    for (let i in auth) {
-      if (auth[i]) {
-        object[i] = auth[i];
-      }
+
+  const hasUsername = _object.auth && _object.auth.username;
+  const hasAuthMechanism = _object.options && _object.options.authMechanism;
+  if (hasUsername || hasAuthMechanism) {
+    object.auth = Object.assign({}, _object.auth);
+    if (object.auth.db) {
+      object.authSource = object.authSource || object.auth.db;
     }
 
-    if (auth.username) {
-      object.auth = auth;
-      object.user = auth.username;
-    }
-
-    if (auth.db) {
-      object.authSource = object.authSource || auth.db;
+    if (object.auth.username) {
+      object.auth.user = object.auth.username;
     }
   }
 
@@ -636,10 +594,12 @@ function validOptions(options) {
       continue;
     }
 
-    if (_validOptions.indexOf(name) === -1 && options.validateOptions) {
-      return new MongoError(`option ${name} is not supported`);
-    } else if (_validOptions.indexOf(name) === -1) {
-      console.warn(`the options [${name}] is not supported`);
+    if (_validOptions.indexOf(name) === -1) {
+      if (options.validateOptions) {
+        return new MongoError(`option ${name} is not supported`);
+      } else {
+        console.warn(`the options [${name}] is not supported`);
+      }
     }
 
     if (legacyOptionNames.indexOf(name) !== -1) {
@@ -651,4 +611,78 @@ function validOptions(options) {
   }
 }
 
-module.exports = { connectOp, logout, validOptions };
+const VALID_AUTH_MECHANISMS = new Set([
+  'DEFAULT',
+  'MONGODB-CR',
+  'PLAIN',
+  'MONGODB-X509',
+  'SCRAM-SHA-1',
+  'SCRAM-SHA-256',
+  'GSSAPI'
+]);
+
+const AUTH_MECHANISM_INTERNAL_MAP = {
+  DEFAULT: 'default',
+  'MONGODB-CR': 'mongocr',
+  PLAIN: 'plain',
+  'MONGODB-X509': 'x509',
+  'SCRAM-SHA-1': 'scram-sha-1',
+  'SCRAM-SHA-256': 'scram-sha-256'
+};
+
+function generateCredentials(client, username, password, options) {
+  options = Object.assign({}, options);
+
+  // the default db to authenticate against is 'self'
+  // if authenticate is called from a retry context, it may be another one, like admin
+  const source = options.authSource || options.authdb || options.dbName;
+
+  // authMechanism
+  const authMechanismRaw = options.authMechanism || 'DEFAULT';
+  const authMechanism = authMechanismRaw.toUpperCase();
+
+  if (!VALID_AUTH_MECHANISMS.has(authMechanism)) {
+    throw MongoError.create({
+      message: `authentication mechanism ${authMechanismRaw} not supported', options.authMechanism`,
+      driver: true
+    });
+  }
+
+  if (authMechanism === 'GSSAPI') {
+    return new MongoCredentials({
+      mechanism: process.platform === 'win32' ? 'sspi' : 'gssapi',
+      mechanismProperties: options,
+      source,
+      username,
+      password
+    });
+  }
+
+  return new MongoCredentials({
+    mechanism: AUTH_MECHANISM_INTERNAL_MAP[authMechanism],
+    source,
+    username,
+    password
+  });
+}
+
+function closeOperation(client, force, callback) {
+  const completeClose = err => {
+    client.emit('close', client);
+    for (const name in client.s.dbCache) {
+      client.s.dbCache[name].emit('close', client);
+    }
+
+    client.removeAllListeners('close');
+    callback(err, null);
+  };
+
+  if (client.topology == null) {
+    completeClose();
+    return;
+  }
+
+  client.topology.close(force, completeClose);
+}
+
+module.exports = { connectOp, validOptions, closeOperation };
